@@ -17,28 +17,17 @@ from posixpath import normpath
 
 import six
 from six.moves import urllib
-from six.moves.urllib.parse import quote, unquote, quote_plus
+from six.moves.urllib.parse import quote, unquote
+try:
+    from icecream import ic
+except ImportError:  # Graceful fallback if IceCream isn't installed.
+    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)
 
 from .omdict1D import omdict1D
-from .compat import basestring, UnicodeMixin
+from .compat import string_types, UnicodeMixin
 from .common import (
     callable_attr, is_iterable_but_not_string, absent as _absent)
 
-#
-# TODO(grun): Subclass Path, PathCompositionInterface, Query, and
-# QueryCompositionInterface into two subclasses each - one for the URL
-# and one for the Fragment.
-#
-# Subclasses will clean up the code because the valid encodings are
-# different between a URL Path and a Fragment Path and a URL Query and a
-# Fragment Query.
-#
-# For example, '?' and '#' don't need to be encoded in Fragment Path
-# segments but must be encoded in URL Path segments.
-#
-# Similarly, '#' doesn't need to be encoded in Fragment Query keys and
-# values, but must be encoded in URL Query keys and values.
-#
 
 # Map of various URL schemes to their default ports. Scheme strings are
 # lowercase.
@@ -73,13 +62,8 @@ def utf8(o, default=_absent):
         return o if default is _absent else default
 
 
-def callable_attr(obj, attr):
-    return hasattr(obj, attr) and callable(getattr(obj, attr))
-
-
-def non_text_iterable(value):
-    b = callable_attr(value, '__iter__') and not isinstance(value, basestring)
-    return b
+def non_string_iterable(o):
+    return callable_attr(o, '__iter__') and not isinstance(o, string_types)
 
 
 # TODO(grun): Support IDNA2008 via the third party idna module. See
@@ -222,6 +206,12 @@ def set_scheme(url, scheme):
         return '%s:%s' % (scheme, after_scheme)
 
 
+# 'netloc' in Python parlance, 'authority' in RFC 3986 parlance.
+def has_netloc(url):
+    scheme = get_scheme(url)
+    return url.startswith('//' if scheme is None else scheme + '://')
+
+
 def urlsplit(url):
     """
     Parameters:
@@ -261,6 +251,7 @@ def urlsplit(url):
         netloc = None
 
     scheme = original_scheme
+
     return urllib.parse.SplitResult(scheme, netloc, path, query, fragment)
 
 
@@ -272,16 +263,30 @@ def urljoin(base, url):
 
     Returns: The resultant URL from joining <base> and <url>.
     """
-    if not base:
-        return url
+    base_scheme = get_scheme(base) if has_netloc(base) else None
+    url_scheme = get_scheme(url) if has_netloc(url) else None
 
-    base_scheme = get_scheme(base)
-    url_scheme = get_scheme(url)
+    if base_scheme is not None:
+        # For consistent URL joining, switch the base URL's scheme to
+        # 'http'. urllib.parse.urljoin() behaves differently depending on the
+        # scheme. E.g.
+        #
+        #   >>> urllib.parse.urljoin('http://google.com/', 'hi')
+        #   'http://google.com/hi'
+        #
+        # vs
+        #
+        #   >>> urllib.parse.urljoin('asdf://google.com/', 'hi')
+        #   'hi'
+        root = set_scheme(base, 'http')
+    else:
+        root = base
 
-    http_base = set_scheme(base, 'http')
-    joined = urllib.parse.urljoin(http_base, url)
-    if not url_scheme:
-        joined = set_scheme(joined, base_scheme)
+    joined = urllib.parse.urljoin(root, url)
+
+    new_scheme = url_scheme if url_scheme is not None else base_scheme
+    if new_scheme is not None and has_netloc(joined):
+        joined = set_scheme(joined, new_scheme)
 
     return joined
 
@@ -305,6 +310,7 @@ def join_path_segments(*args):
     Returns: A list containing the joined path segments.
     """
     finals = []
+
     for segments in args:
         if not segments or segments == ['']:
             continue
@@ -319,6 +325,7 @@ def join_path_segments(*args):
             elif finals[-1] != '' and segments[0] == '' and len(segments) > 1:
                 segments = segments[1:]
             finals.extend(segments)
+
     return finals
 
 
@@ -398,6 +405,13 @@ class Path(object):
         take such strings, like load(), add(), set(), remove(), etc.
     """
 
+    # RFC 3986:
+    #
+    # segment       = *pchar
+    # pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
+    # unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
+    # sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
+    #                     / "*" / "+" / "," / ";" / "="
     SAFE_SEGMENT_CHARS = ":@-._~!$&'()*+,;="
 
     def __init__(self, path='', force_absolute=lambda _: False, strict=False):
@@ -457,6 +471,7 @@ class Path(object):
             segments.insert(0, '')
 
         self.load(join_path_segments(segments, newsegments))
+
         return self
 
     def set(self, path):
@@ -473,6 +488,7 @@ class Path(object):
                 segments = self._segments_from_path(path)
             base = ([''] if self.isabsolute else []) + self.segments
             self.load(remove_path_segments(base, segments))
+
         return self
 
     def normalize(self):
@@ -486,6 +502,7 @@ class Path(object):
             if normalized.startswith('//'):  # http://bugs.python.org/636648
                 normalized = '/' + normalized.lstrip('/')
             self.load(normalized)
+
         return self
 
     def asdict(self):
@@ -534,6 +551,9 @@ class Path(object):
         """
         return not self.isdir
 
+    def __truediv__(self, path):
+        return self.add(path)
+
     def __eq__(self, other):
         return str(self) == str(other)
 
@@ -579,10 +599,9 @@ class Path(object):
             segments.append(utf8(segment))
         del segment
 
-        # In Python 3, utf8() returns Bytes objects that must be decoded
-        # into strings before they can be passed to
-        # six.moves.urllib.parse.unquote(). In Python 2, utf8() returns
-        # strings that can be passed directly to urllib.unquote().
+        # In Python 3, utf8() returns Bytes objects that must be decoded into
+        # strings before they can be passed to unquote(). In Python 2, utf8()
+        # returns strings that can be passed directly to urllib.unquote().
         segments = [
             segment.decode('utf8')
             if isinstance(segment, bytes) and not isinstance(segment, str)
@@ -794,10 +813,6 @@ class Query(object):
         take such strings, like load(), add(), set(), remove(), etc.
     """
 
-    SAFE_KEY_CHARS = "/?:@-._~!$'()*,"
-    SAFE_VALUE_CHARS_NON_EMPTY_KEY = SAFE_KEY_CHARS
-    SAFE_VALUE_CHARS_EMPTY_KEY = SAFE_KEY_CHARS + '='
-
     def __init__(self, query='', strict=False):
         self.strict = strict
 
@@ -845,16 +860,17 @@ class Query(object):
         if callable_attr(query, 'items'):
             items = self._items(query)
         # List of keys or items to remove.
-        elif non_text_iterable(query):
+        elif non_string_iterable(query):
             items = query
 
         for item in items:
-            if non_text_iterable(item) and len(item) == 2:
+            if non_string_iterable(item) and len(item) == 2:
                 key, value = item
                 self.params.popvalue(key, value, None)
             else:
                 key = item
                 self.params.pop(key, None)
+
         return self
 
     @property
@@ -892,17 +908,14 @@ class Query(object):
 
         pairs = []
         sixurl = urllib.parse  # six.moves.urllib.parse
-        quote_func = sixurl.quote_plus if quote_plus else sixurl.quote
+        quote_fn = sixurl.quote_plus if quote_plus else sixurl.quote
         for key, value in self.params.iterallitems():
             utf8key = utf8(key, utf8(attemptstr(key)))
             utf8value = utf8(value, utf8(attemptstr(value)))
 
-            quoted_key = quote_func(utf8key, self.SAFE_KEY_CHARS)
-            if not quoted_key:
-                safe_value_chars = self.SAFE_VALUE_CHARS_EMPTY_KEY
-            else:
-                safe_value_chars = self.SAFE_VALUE_CHARS_NON_EMPTY_KEY
-            quoted_value = quote_func(utf8value, safe_value_chars)
+            quoted_key = quote_fn(utf8key)
+            safe_value_chars = '=' if not quoted_key else ''
+            quoted_value = quote_fn(utf8value, safe_value_chars)
 
             if value is None:  # Example: http://sprop.su/?param.
                 pair = quoted_key
@@ -1118,6 +1131,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.add(path)
         if args is not _absent:
             self.query.add(args)
+
         return self
 
     def set(self, path=_absent, args=_absent, separator=_absent):
@@ -1127,6 +1141,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.query.load(args)
         if separator is True or separator is False:
             self.separator = separator
+
         return self
 
     def remove(self, fragment=_absent, path=_absent, args=_absent):
@@ -1136,6 +1151,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.remove(path)
         if args is not _absent:
             self.query.remove(args)
+
         return self
 
     def asdict(self):
@@ -1171,9 +1187,9 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
         if path and (not query or not self.separator):
             path = path.replace('%3F', '?')
 
-        if query and path:
-            return path + ('?' if self.separator else '') + query
-        return path + query
+        separator = '?' if path and query and self.separator else ''
+
+        return path + separator + query
 
     def __repr__(self):
         return "%s('%s')" % (self.__class__.__name__, str(self))
@@ -1282,6 +1298,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         self.path.load(tokens.path)
         self.query.load(tokens.query)
         self.fragment.load(tokens.fragment)
+
         return self
 
     @property
@@ -1413,6 +1430,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
             port = ':%s' % self.port
         origin = '%s://%s%s' % (scheme, host, port)
+
         return origin
 
     @origin.setter
@@ -1482,6 +1500,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             self.query.add(query_params)
         if fragment_path is not _absent or fragment_args is not _absent:
             self.fragment.add(path=fragment_path, args=fragment_args)
+
         return self
 
     def set(self, args=_absent, path=_absent, fragment=_absent, scheme=_absent,
@@ -1720,6 +1739,9 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             'fragment': self.fragment.asdict(),
             }
 
+    def __truediv__(self, path):
+        return self.add(path=path)
+
     def __eq__(self, other):
         try:
             return self.url == other.url
